import React from 'react';
import { defaultTransforms } from './transforms/defaultTransforms.js';
import { DoNotRenderBBCodeError } from './DoNotRenderBBCodeError.js';
import { BBCodeOriginalStackTrace } from './BBCodeOriginalStackTrace.js';
export class Generator {
    errorIfNoTransform;
    transforms;
    constructor(transforms = defaultTransforms, errorIfNoTransform = true) {
        this.errorIfNoTransform = errorIfNoTransform;
        this.transforms = new Map(transforms.map((transform) => [transform.name, transform]));
    }
    joinAdjacentStrings(children) {
        return children.reduce((acc, child) => {
            if (typeof child === 'string' && typeof acc[acc.length - 1] === 'string') {
                acc[acc.length - 1] += child;
            }
            else {
                acc.push(child);
            }
            return acc;
        }, []);
    }
    createUnrenderedNodeWithRenderedChildren(tagNode, children) {
        if (!children || children.length === 0)
            return [tagNode.ogStartTag, tagNode.ogEndTag];
        if (typeof children[0] === 'string') {
            children[0] = tagNode.ogStartTag + children[0];
        }
        else {
            children.unshift(tagNode.ogStartTag);
        }
        if (typeof children[children.length - 1] === 'string') {
            children[children.length - 1] = children[children.length - 1] + tagNode.ogEndTag;
        }
        else {
            children.push(tagNode.ogEndTag);
        }
        return children;
    }
    generateForNode(node, key) {
        switch (node.nodeType) {
            case 2 /* AstNodeType.LinebreakNode */: {
                return React.createElement("br", { key: key });
            }
            case 1 /* AstNodeType.TextNode */: {
                return node.str;
            }
            case 3 /* AstNodeType.TagNode */: {
                const tagName = node.tagName;
                const transform = this.transforms.get(tagName);
                const renderedChildren = transform && transform.skipChildren
                    ? undefined
                    : this.joinAdjacentStrings(node.children.map((child, i) => {
                        const renderedChild = this.generateForNode(child, i);
                        return renderedChild;
                    })).flat();
                if (!transform) {
                    if (this.errorIfNoTransform) {
                        throw new Error(`Unrecognized bbcode ${node.tagName}`);
                    }
                    else {
                        console.warn(`Unrecognized bbcode ${node.tagName}`);
                    }
                    return this.createUnrenderedNodeWithRenderedChildren(node);
                }
                // because error boundaries don't work compile-time smh
                const WrappedComponentFunction = ({ tagNode, children }) => {
                    try {
                        return transform.Component({ tagNode, children });
                    }
                    catch (e) {
                        if (!(e instanceof DoNotRenderBBCodeError)) {
                            if (e instanceof Error && e.stack && e[BBCodeOriginalStackTrace] === undefined) {
                                e[BBCodeOriginalStackTrace] = e.stack;
                                e.stack += (`\n\nTag: [${node.tagName}]\n\n` + // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
                                    `${this.createUnrenderedNodeWithRenderedChildren(node)}` +
                                    `\n\n${node.toString()}`).split('\n').map((line) => line.length ? `    ${line}` : '').join('\n');
                            }
                            throw e;
                        }
                        return this.createUnrenderedNodeWithRenderedChildren(node, renderedChildren);
                    }
                };
                // 😱 readable component names in error traces FTW
                Object.defineProperty(WrappedComponentFunction, 'name', {
                    value: transform.Component.name !== '' && transform.Component.name !== 'Component'
                        ? transform.Component.name
                        : `BBCode_${tagName}`,
                });
                return React.createElement(WrappedComponentFunction, { tagNode: node, key: key }, renderedChildren);
            }
            default: {
                const renderedChildren = node.children.map((child, i) => {
                    const renderedChild = this.generateForNode(child, i);
                    return renderedChild;
                });
                return this.joinAdjacentStrings(renderedChildren);
            }
        }
    }
    generate(root) {
        return React.createElement(React.Fragment, null, this.generateForNode(root, 0));
    }
}
//# sourceMappingURL=Generator.js.map