/**

Haven't formally verified this grammar but it should be LL(2)

The root's intermediate state has StartTag/EndTag because it's easier to first parse them as independant nodes
than to parse a StartTag and find the matching EndTag since we can only lookahead by 1 token

Trying to lookahead by 4 tokens after each advancement to determine the end of the sub-root will greatly affect performance
    1 "["
    2 "/"
    3 "LABEL"
    4 "]"

---

Root <- (Text | Linebreak | Tag)*

Text <-
| {XSS Characters}.
| STR.

Linebreak <-
| LINEBREAK.

Tag      <- StartTag Root EndTag
StartTag <- L_BRACKET           Text Attr* R_BRACKET
EndTag   <- L_BRACKET BACKSLASH Text       R_BRACKET

Attr <-
| STR EQUALS STR
| EQUALS STR
| STR

*/
import { nodeIsType } from './nodeIsType.js';
export function nodeTypeToString(nodeType) {
    switch (nodeType) {
        case 0 /* AstNodeType.RootNode */: return 'RootNode';
        case 1 /* AstNodeType.TextNode */: return 'TextNode';
        case 2 /* AstNodeType.LinebreakNode */: return 'LinebreakNode';
        case 3 /* AstNodeType.TagNode */: return 'TagNode';
        case 4 /* AstNodeType.StartTagNode */: return 'StartTagNode';
        case 5 /* AstNodeType.EndTagNode */: return 'EndTagNode';
        case 6 /* AstNodeType.AttrNode */: return 'AttrNode';
    }
}
export class AstNodeBase {
    // eslint-disable-next-line no-use-before-define
    children;
    constructor(children = []) {
        this.children = children;
    }
    addChild(node) {
        this.children.push(node);
    }
    isValid() {
        for (const child of this.children) {
            if (!child.isValid()) {
                return false;
            }
        }
        return true;
    }
    toShortString() {
        return nodeTypeToString(this.nodeType);
    }
    // For debugging purposes only
    // Pretty-prints AST
    toString(depth = 0) {
        let s = ' '.repeat(depth * 2) + this.toShortString();
        for (const child of this.children) {
            s += '\n' + child.toString(depth + 1);
        }
        return s;
    }
}
// ----------------------------------------------------------------------------
// Root
// ----------------------------------------------------------------------------
export class RootNode extends AstNodeBase {
    nodeType = 0 /* AstNodeType.RootNode */;
    isValid() {
        for (const child of this.children) {
            if (child.nodeType !== 3 /* AstNodeType.TagNode */ &&
                child.nodeType !== 1 /* AstNodeType.TextNode */ &&
                child.nodeType !== 2 /* AstNodeType.LinebreakNode */) {
                return false;
            }
        }
        return super.isValid() && this.children.length > 0;
    }
}
// ----------------------------------------------------------------------------
// Text
// ----------------------------------------------------------------------------
export class TextNode extends AstNodeBase {
    nodeType = 1 /* AstNodeType.TextNode */;
    str;
    constructor(str) {
        super();
        this.str = str;
    }
    isValid() {
        return super.isValid() && this.children.length === 0;
    }
    toShortString() {
        return `${super.toShortString()} "${this.str}"`;
    }
}
export class LinebreakNode extends AstNodeBase {
    nodeType = 2 /* AstNodeType.LinebreakNode */;
    toShortString() {
        return `${super.toShortString()} "\\n"`;
    }
}
// ----------------------------------------------------------------------------
// Tag
// ----------------------------------------------------------------------------
export class StartTagNode extends AstNodeBase {
    nodeType = 4 /* AstNodeType.StartTagNode */;
    tagName;
    ogTag;
    constructor(tagName, ogTag, attrNodes = []) {
        super(attrNodes);
        this.tagName = tagName.toLowerCase();
        this.ogTag = ogTag;
    }
    isValid() {
        for (const child of this.children) {
            if (child.nodeType !== 6 /* AstNodeType.AttrNode */) {
                return false;
            }
        }
        return super.isValid();
    }
    toShortString() {
        return `${super.toShortString()} ${this.ogTag}`;
    }
}
export class EndTagNode extends AstNodeBase {
    nodeType = 5 /* AstNodeType.EndTagNode */;
    tagName;
    ogTag;
    constructor(tagName, ogTag) {
        super();
        this.tagName = tagName;
        this.ogTag = ogTag;
    }
    isValid() {
        return super.isValid() && this.children.length === 0;
    }
    toShortString() {
        return `${super.toShortString()} ${this.ogTag}`;
    }
}
export class TagNode extends AstNodeBase {
    nodeType = 3 /* AstNodeType.TagNode */;
    _startTag;
    _endTag;
    constructor(startTag, endTag) {
        super();
        this._startTag = startTag;
        this._endTag = endTag;
    }
    get tagName() {
        return this._startTag.tagName;
    }
    get attributes() {
        return this._startTag.children;
    }
    getAttributesByName() {
        const attrs = {};
        for (const attrNode of this.attributes) {
            attrs[attrNode.key] = attrNode.val;
        }
        return attrs;
    }
    get ogStartTag() {
        return this._startTag.ogTag;
    }
    get ogEndTag() {
        if (!this._endTag) {
            return '';
        }
        if (nodeIsType(this._endTag, 2 /* AstNodeType.LinebreakNode */)) {
            return '\n';
        }
        else {
            return this._endTag.ogTag;
        }
    }
    isValid() {
        if (this._endTag && nodeIsType(this._endTag, 5 /* AstNodeType.EndTagNode */) && this._startTag.tagName !== this._endTag.tagName) {
            return false;
        }
        if (this.children.length === 1 && this.children[0].nodeType !== 0 /* AstNodeType.RootNode */) {
            return false;
        }
        if (this.children.length > 2) {
            return false;
        }
        return super.isValid() && this._startTag.isValid() && (this._endTag?.isValid() ?? true);
    }
    toString(depth = 0) {
        let s = ' '.repeat(depth * 2) + this.toShortString() + ` [${this.tagName}]`;
        for (const attrNode of this._startTag.children) {
            s += '\n' + attrNode.toString(depth + 1);
        }
        for (const child of this.children) {
            s += '\n' + child.toString(depth + 1);
        }
        return s;
    }
    /**
     * Gets the text of the immediate descendant of the current TagNode
     *
     * For example, would return "https://en.wikipedia.org/wiki/Markdown" for the following:
     * ```bbcode
     * [url]https://en.wikipedia.org/wiki/Markdown[/url]
     * ```
     */
    getTagImmediateText() {
        if (this.children.length !== 1) {
            return undefined;
        }
        const child = this.children[0];
        if (!nodeIsType(child, 0 /* AstNodeType.RootNode */)) {
            return undefined;
        }
        if (child.children.length !== 1) {
            return undefined;
        }
        const textNode = child.children[0];
        if (!nodeIsType(textNode, 1 /* AstNodeType.TextNode */)) {
            return undefined;
        }
        return textNode.str;
    }
    /**
     * Gets the value of the immediate attribute of the current TagNode
     *
     * For example, would return "https://en.wikipedia.org/wiki/Markdown" for the following:
     * ```bbcode
     * [url=https://en.wikipedia.org/wiki/Markdown] Wikipedia's page on Markdown [/url]
     * ```
     */
    getTagImmediateAttrVal() {
        if (this.attributes.length !== 1) {
            return undefined;
        }
        const attrNode = this.attributes[0];
        return attrNode.val;
    }
}
// ----------------------------------------------------------------------------
// Attr
// ----------------------------------------------------------------------------
export class AttrNode extends AstNodeBase {
    nodeType = 6 /* AstNodeType.AttrNode */;
    static DEFAULT_KEY = 'default';
    get key() {
        switch (this.children.length) {
            case 1: {
                return AttrNode.DEFAULT_KEY;
            }
            case 2: {
                if (!nodeIsType(this.children[0], 1 /* AstNodeType.TextNode */)) {
                    throw new Error('Invalid TextNode');
                }
                return this.children[0].str.trim();
            }
        }
        throw new Error('Invalid AttrNode');
    }
    get val() {
        switch (this.children.length) {
            case 1: {
                if (!nodeIsType(this.children[0], 1 /* AstNodeType.TextNode */)) {
                    throw new Error('Invalid TextNode');
                }
                return this.children[0].str.trim();
            }
            case 2: {
                if (!nodeIsType(this.children[1], 1 /* AstNodeType.TextNode */)) {
                    throw new Error('Invalid TextNode');
                }
                return this.children[1].str.trim();
            }
        }
        throw new Error('Invalid AttrNode');
    }
    isValid() {
        return super.isValid() && (this.children.length >= 1 && this.children.length <= 2);
    }
    toShortString() {
        let s = super.toShortString();
        switch (this.children.length) {
            case 1: {
                s += ` VAL="${this.val}"`;
                break;
            }
            case 2: {
                s += ` KEY="${this.key}" VAL="${this.val}"`;
                break;
            }
        }
        return s;
    }
}
//# sourceMappingURL=AstNode.js.map